﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Joseph Helm and Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Encodings.PackingEncoding.MultiComponentVector {
  [Item("Bin Based Multi Component Vector Crossover", "An operator which crosses two MultiComponentVector representations. The applied strategy is based on the RBX crossover proposed by Potvin.")]
  [StorableClass]
  public class BinBasedMultiComponentVectorCrossover :  MultiComponentVectorCrossover {
    [StorableConstructor]
    protected BinBasedMultiComponentVectorCrossover(bool deserializing) : base(deserializing) { }
    protected BinBasedMultiComponentVectorCrossover(BinBasedMultiComponentVectorCrossover original, Cloner cloner) : base(original, cloner) { }
    public BinBasedMultiComponentVectorCrossover()
      : base() {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new BinBasedMultiComponentVectorCrossover(this, cloner);
    }

    public override MultiComponentVectorEncoding Cross(IRandom random, MultiComponentVectorEncoding parent1, MultiComponentVectorEncoding parent2) {
      MultiComponentVectorEncoding child = new MultiComponentVectorEncoding ();

      int nrOfItems = parent1.NrOfItems;

      bool[] itemAlreadyAssigned = new bool[nrOfItems];
      int nrOfBins = parent1.NrOfBins > parent2.NrOfBins ? parent2.NrOfBins : parent1.NrOfBins;

      //Add one bin from parent1
      int swappedBinNr = random.Next(nrOfBins);
      var swappedBin = new ItemList<PackingInformation>();
      for (int i = 0; i < parent1.PackingInformations[swappedBinNr].Count; i++) {
        PackingInformation pi = new PackingInformation(parent1.PackingInformations[swappedBinNr][i]);
        if (!itemAlreadyAssigned[pi.ItemID]) {
          itemAlreadyAssigned[pi.ItemID] = true;
          swappedBin.Add(new PackingInformation(pi));
        }
      }
      child.PackingInformations[swappedBinNr] = swappedBin;

      //Get the remaining bins from parent2
      for (int binNr = 0; binNr < nrOfBins; binNr++) {
        if (binNr != swappedBinNr) {
          var newBin = new ItemList<PackingInformation>();
          var currentParent = parent2;
          int nrOfItemsInBin = currentParent.PackingInformations[binNr].Count;
          for (int i = 0; i < nrOfItemsInBin; i++) {
            PackingInformation pi = new PackingInformation(currentParent.PackingInformations[binNr][i]);
            if (!itemAlreadyAssigned[pi.ItemID]) {
              itemAlreadyAssigned[pi.ItemID] = true;
              newBin.Add(new PackingInformation(pi));
            }
          }
          child.PackingInformations[binNr] = newBin;
        }
      }

      //Pack still remaining items in bin#0
      for (int itemID = 0; itemID < nrOfItems; itemID++) {
        if (!itemAlreadyAssigned[itemID])
          child.PackingInformations[0].Add(new PackingInformation(itemID, random.Next(2) == 0 ? true : false));
      }                

      return child;
    }       
  }
}
